/* Copyright (c) 2002-2020 CrystalClear Software, Inc.
 * Use, modification and distribution is subject to the 
 * Boost Software License, Version 1.0. (See accompanying
 * file LICENSE_1_0.txt or http://www.boost.org/LICENSE_1_0.txt)
 * Author: Jeff Garland, Bart Garst
 * $Date$
 */


// define BOOST_DATE_TIME_SOURCE so that <boost/date_time/config.hpp> knows
// the library is being built (possibly exporting rather than importing code)
#ifndef BOOST_DATE_TIME_SOURCE
#define BOOST_DATE_TIME_SOURCE
#endif

#include <boost/date_time/compiler_config.hpp>

namespace boost {
namespace gregorian {


BOOST_DATE_TIME_DECL void date_time_dummy_exported_function()
{}

} } //namespace boost::gregorian






